<h3>Kombu</h3>
<p>
  Kombu is an AMQP messaging framework for Python.
</p>
<p class="logo"><a href="{{ pathto(master_doc) }}">
  <img width="128" height="128" class="logo" src="http://cloud.github.com/downloads/ask/kombu/kombusmall.jpg" alt="Logo"/>
</a></p>
